Attribute VB_Name = "modLivelliPS"
'==============================================================
' Descrizione.....: Subroutines DELLIV ( F. Languasco).
' Nome dei Files..: modLivelliPS.bas, modDelaunay2D.bas
'                   modGeometria.bas, modUtilita.bas
' Data............: 17/07/2004
' Versione........: 1.0 a 32 bits.
' Sistema.........: VB6 (SP5) sotto Windows NT (SP6).
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://www.flanguasco.org
'                   http://members.xoom.virgilio.it/flanguasco
'==============================================================
'
'   Subroutines DELLIV per il tracciamento delle curve di livello
'   di superfici definite da punti sparpagliati.  Le superfici
'   sono considerate a faccie triangolari, con vertici dei triangoli
'   dati dalla triangolazione di Delaunay dei punti con coordinate
'   contenute nei vettori XD(), YD() e ZD().  Le linee di livello L
'   sono calcolate come intersezione dei triangoli di Delaunay con
'   i piani di equazione z = ZLin(L).dLinVal.
'
'   Routines disponibili:
'
'    DELLIV_Lines picPage As PictureBox, XD#(), YD#(), ZD#(), _
'                 ZLin() As LineaLivello_Type, [Msg$], _
'                 [DisegnaTriangoli As Boolean = False], _
'                 [DisegnaPerimetro As Boolean = False], _
'                 [fStop As Boolean]
'
'    Disegna, con linee formate da segmenti successivi, le curve di
'    livello della superficie
'    Parametri:
'     picPage:          PictureBox su cui si vuole disegnare.
'     XD(), YD(), ZD(): vettori con le coordinate dei punti della superficie.
'     ZLin(1 to NLiv):  struttura contenente i parametri delle linee di livello
'                       da disegnare (vedi sotto).
'     Msg$:             stringa con messaggi di errore; Msg$ = "" in caso di successo.
'     DisegnaTriangoli: se True disegna anche la triangolazione di Delaunay
'     DisegnaPerimetro: se True disegna il "convex hull" dei punti dati sul piano X-Y.
'     fStop:            da mettere a True, nella routine chiamante, per interrompere
'                       il calcolo delle linee di livello.
'
'    DELLIV_Solid picPage As PictureBox, XD#(), YD#(), ZD#(), _
'                 ZLin() As LineaLivello_Type, [Msg$], _
'                 [DisegnaTriangoli As Boolean = False], _
'                 [DisegnaPerimetro As Boolean = False], _
'                 [fStop As Boolean]
'
'    Disegna, con poligoni colorati, le aree della superficie a livello
'    compreso tra i piani di equazione z = ZLin(L).dLinVal e z = ZLin(L + 1).dLinVal,
'    con 1 <= L < NLiv.
'    Parametri:
'     come per DELLIV_Lines.
'
Option Explicit
'
Private NLiv&   ' N di livelli.
'
' Struttura delle informazioni relative
' alle linee di livello da disegnare:
Public Type LineaLivello_Type
    dLinVal As Double      ' Valore    della linea di livello.
    lLinCol As Long        ' Colore      "     "    "    " (default = vbBlack).
    lLinSps As Long        ' Spessore    "     "    "    " (default = 1 [Pixel]).
    sLblTes As String      ' Testo     dell' etichetta.
    lLblCol As Long        ' Colore      "      "  (default = vbBlack).
    lLblFSz As Long        ' Font size   "      "  (default = 8 [Points]).
End Type
'
' Struttura di appoggio per scrivere
' le etichette dei livelli:
Private Type zLbl_Type
    LblW As Double  ' Larghezza dell' etichetta.
    LblH As Double  ' Altezza   dell' etichetta.
    X1 As Single    ' Posizioni degli angoli
    X2 As Single    ' dell' area impegnata
    Y1 As Single    ' dall' etichetta.
    Y2 As Single    '
    OK As Boolean   ' Linea di livello etichettata.
End Type
Private zLbl() As zLbl_Type

Public Sub DELLIV_Solid(ByVal picPage As PictureBox, XD#(), YD#(), ZD#(), _
    ZLin() As LineaLivello_Type, _
    Optional ByRef Msg$, _
    Optional ByVal DisegnaTriangoli As Boolean = False, _
    Optional ByVal DisegnaPerimetro As Boolean = False, _
    Optional ByRef fStop As Boolean = False)
'
'
    Dim I&, J&, K&, N&, L&, ND&, NT&, num_int&, IERR&
    Dim XI#(), YI#(), ZI#()
'
    On Error GoTo DELLIV_Solid_ERR
'
    NLiv = UBound(ZLin)
    If NLiv <= 1 Then Err.Raise 1001, "DELLIV_Lines", "Errore nei livelli: NLiv = " & NLiv
    For L = 2 To NLiv
        If ZLin(L).dLinVal <= ZLin(L - 1).dLinVal Then
            Err.Raise 1001, "DELLIV_Lines", "Errore nei livelli:" & vbNewLine _
                    & "ZLin(" & L - 1 & ").dLinVal >= ZLin(" & L & ").dLinVal"
            Exit For
        End If
    Next L
'
    ND = UBound(XD)
    ReDim Vcl#(1 To 2, 1 To ND), Ind&(1 To ND)
    ReDim TIL&(1 To 3, 1 To ND * 2), TNBR&(1 To 3, 1 To ND * 2)
'
    For N = 1 To ND
        ' Organizzo le coordinate dei vertici come
        ' richiesto dalle routines di modDelaunay2D:
        Vcl(1, N) = XD(N)
        Vcl(2, N) = YD(N)
        Ind(N) = N  ' Triangolazione di tutti i vertici contenuti in VCL().
    Next N
'
    ' Calcolo, sul piano X-Y, la triangolazione
    ' dei punti dati:
    'DTRIS2 ND, VCL(), Ind(), NT, TIL(), TNBR(), IERR
    DTRIW2 ND, Vcl(), Ind(), NT, TIL(), TNBR(), IERR
'
    If IERR <> 0 Then
        Err.Raise 1001, "DELLIV_Solid", "Errore di triangolazione IERR = " & IERR _
                & vbNewLine & "(Vedere tabella degli errori)"
    End If
'
    Dim NInt&, IntX1#(), IntY1#(), IntX2#(), IntY2#(), IntZ#()
    For N = 1 To NT
        NInt = 0
        For L = 1 To NLiv
            ' Trovo le intersezioni del triangolo N
            ' con il piano a z = ZLin(L).dLinVal:
            plane_imp_triangle_int_3d XD(TIL(1, N)), YD(TIL(1, N)), ZD(TIL(1, N)), _
                                      XD(TIL(2, N)), YD(TIL(2, N)), ZD(TIL(2, N)), _
                                      XD(TIL(3, N)), YD(TIL(3, N)), ZD(TIL(3, N)), _
                                      0#, 0#, 1#, -ZLin(L).dLinVal, _
                                      num_int, XI(), YI(), ZI()
'
            If num_int = 2 Then
                ' Aggiungo le coordinate delle nuove intersezioni trovate
                ' alle liste IntX1(), IntY1(), IntX2(), IntY2(), IntZ():
                    NInt = NInt + 1
                    ReDim Preserve IntX1(1 To NInt), IntY1(1 To NInt)
                    ReDim Preserve IntX2(1 To NInt), IntY2(1 To NInt)
                    ReDim Preserve IntZ(1 To NInt)
                    IntX1(NInt) = XI(1)
                    IntY1(NInt) = YI(1)
                    IntX2(NInt) = XI(2)
                    IntY2(NInt) = YI(2)
                    IntZ(NInt) = ZLin(L).dLinVal
            End If
        Next L
'
        If NInt = 0 Then
            ' Non ci sono intersezioni e quindi i vertici
            ' del triangolo N sono compresi fra due livelli:
            ReDim XV#(1 To 3), YV#(1 To 3), ZV#(1 To 3)
            For I = 1 To 3
                XV(I) = XD(TIL(I, N))
                YV(I) = YD(TIL(I, N))
                ZV(I) = ZD(TIL(I, N))
            Next I
            DisegnaTessera picPage, XV#(), YV#(), ZV#(), ZLin()
'
        Else
            ' Gestisco, per prime, le intersezioni con il primo livello
            ' ed i vertici del triangolo N sotto il primo livello:
            J = 2
            ReDim XV#(1 To J), YV#(1 To J), ZV#(1 To J)
            XV(1) = IntX1(1): YV(1) = IntY1(1): ZV(1) = IntZ(1)
            XV(2) = IntX2(1): YV(2) = IntY2(1): ZV(2) = IntZ(1)
            For I = 1 To 3
                If ZD(TIL(I, N)) < IntZ(1) Then
                    J = J + 1
                    ReDim Preserve XV#(1 To J), YV#(1 To J), ZV#(1 To J)
                    XV(J) = XD(TIL(I, N))
                    YV(J) = YD(TIL(I, N))
                    ZV(J) = ZD(TIL(I, N))
                End If
            Next I
            If J > 2 Then DisegnaTessera picPage, XV#(), YV#(), ZV#(), ZLin()
'
            ' Gestisco le intersezioni con i livelli K e K + 1
            ' ed i vertici del triangolo compresi fra i livelli
            ' K e K + 1:
            For K = 1 To NInt - 1
                J = 4
                ReDim XV#(1 To J), YV#(1 To J), ZV#(1 To J)
                XV(1) = IntX1(K): YV(1) = IntY1(K): ZV(1) = IntZ(K)
                XV(2) = IntX2(K): YV(2) = IntY2(K): ZV(2) = IntZ(K)
                XV(3) = IntX1(K + 1): YV(3) = IntY1(K + 1): ZV(3) = IntZ(K + 1)
                XV(4) = IntX2(K + 1): YV(4) = IntY2(K + 1): ZV(4) = IntZ(K + 1)
                For I = 1 To 3
                    If IntZ(K) < ZD(TIL(I, N)) And ZD(TIL(I, N)) < IntZ(K + 1) Then
                        J = J + 1
                        ReDim Preserve XV#(1 To J), YV#(1 To J), ZV#(1 To J)
                        XV(J) = XD(TIL(I, N))
                        YV(J) = YD(TIL(I, N))
                        ZV(J) = ZD(TIL(I, N))
                    End If
                Next I
                DisegnaTessera picPage, XV#(), YV#(), ZV#(), ZLin()
            Next K
'
            ' Gestisco, infine, le intersezioni con l' ultimo livello
            ' ed i vertici del triangolo N sopra l' ultimo livello:
            J = 2
            ReDim XV#(1 To J), YV#(1 To J), ZV#(1 To J)
            XV(1) = IntX1(NInt): YV(1) = IntY1(NInt): ZV(1) = IntZ(NInt)
            XV(2) = IntX2(NInt): YV(2) = IntY2(NInt): ZV(2) = IntZ(NInt)
            For I = 1 To 3
                If IntZ(NInt) < ZD(TIL(I, N)) Then
                    J = J + 1
                    ReDim Preserve XV#(1 To J), YV#(1 To J), ZV#(1 To J)
                    XV(J) = XD(TIL(I, N))
                    YV(J) = YD(TIL(I, N))
                    ZV(J) = ZD(TIL(I, N))
                End If
            Next I
            If J > 2 Then DisegnaTessera picPage, XV#(), YV#(), ZV#(), ZLin()
        End If
'
        If fStop Then Exit Sub
    Next N
'
    If DisegnaTriangoli Then
        ' Disegno i triangoli:
        For N = 1 To NT
            picPage.Line (XD(TIL(1, N)), YD(TIL(1, N))) _
                        -(XD(TIL(2, N)), YD(TIL(2, N))), vbWhite
            picPage.Line (XD(TIL(2, N)), YD(TIL(2, N))) _
                        -(XD(TIL(3, N)), YD(TIL(3, N))), vbWhite
            picPage.Line (XD(TIL(3, N)), YD(TIL(3, N))) _
                        -(XD(TIL(1, N)), YD(TIL(1, N))), vbWhite
        Next N
    End If
'
    If DisegnaPerimetro Then
        ' Disegno il perimetro:
        Dim LINK&, I1&
        For N = 1 To NT
            For I = 1 To 3
                LINK = TNBR(I, N)
                I1 = I + 1: If I1 > 3 Then I1 = 1
                If LINK < 0 Then
                    picPage.Line (XD(TIL(I, N)), YD(TIL(I, N))) _
                                -(XD(TIL(I1, N)), YD(TIL(I1, N))), vbWhite
                End If
            Next I
        Next N
    End If
'
'
DELLIV_Solid_ERR:
    Msg$ = IIf(Err = 0, "", Err.Description)
'
'
'
End Sub
Public Sub DELLIV_Lines(ByVal picPage As PictureBox, XD#(), YD#(), ZD#(), _
    ZLin() As LineaLivello_Type, _
    Optional ByRef Msg$, _
    Optional ByVal DisegnaTriangoli As Boolean = False, _
    Optional ByVal DisegnaPerimetro As Boolean = False, _
    Optional ByRef fStop As Boolean = False)
'
'
    Dim I&, N&, L&, ND&, NT&, num_int&, IERR&
    Dim XI#(), YI#(), ZI#()
'
    ' Variabili per le etichette dei livelli:
    Dim PageFN$, PageFS&, PageDW&, PageFC&
    Dim fzLbl As Boolean, PageFB As Boolean, PageFT As Boolean
'
    On Error GoTo DELLIV_Lines_ERR
'
    NLiv = UBound(ZLin)
    If NLiv <= 1 Then Err.Raise 1001, "DELLIV_Lines", "Errore nei livelli: NLiv = " & NLiv
    For L = 2 To NLiv
        If ZLin(L).dLinVal <= ZLin(L - 1).dLinVal Then
            Err.Raise 1001, "DELLIV_Lines", "Errore nei livelli:" & vbNewLine _
                    & "ZLin(" & L - 1 & ").dLinVal >= ZLin(" & L & ").dLinVal"
            Exit For
        End If
    Next L
    ReDim zLbl(1 To NLiv) As zLbl_Type
'
    ND = UBound(XD)
    ReDim Vcl#(1 To 2, 1 To ND), Ind&(1 To ND)
    ReDim TIL&(1 To 3, 1 To ND * 2), TNBR&(1 To 3, 1 To ND * 2)
'
    For N = 1 To ND
        ' Organizzo le coordinate dei vertici come
        ' richiesto dalle routines di modDelaunay2D:
        Vcl(1, N) = XD(N)
        Vcl(2, N) = YD(N)
        Ind(N) = N  ' Triangolazione di tutti i vertici contenuti in VCL().
    Next N
'
    ' Calcolo, sul piano X-Y, la triangolazione
    ' dei punti dati:
    'DTRIS2 ND, VCL(), Ind(), NT, TIL(), TNBR(), IERR
    DTRIW2 ND, Vcl(), Ind(), NT, TIL(), TNBR(), IERR
'
    If IERR <> 0 Then
        Err.Raise 1001, "DELLIV_Lines", "Errore di triangolazione IERR = " & IERR _
                & vbNewLine & "(Vedere tabella degli errori)"
    End If
'
    ' Calcolo la lunghezza in [Caratteri] delle etichette
    ' e ne richiedo, eventualmente, la visualizzazione:
    For L = 1 To NLiv
        zLbl(L).LblW = Len(Trim(ZLin(L).sLblTes))
        If zLbl(L).LblW > 0 Then
            zLbl(L).LblW = zLbl(L).LblW + 2
            fzLbl = True
        End If
    Next L
'
    PageDW = picPage.DrawWidth
    If fzLbl Then   ' Si vogliono scrivere i valori dei livelli.
        ' Salvo le impostazioni correnti,
        ' assegno quelle nuove e calcolo
        ' i parametri di posizione:
        PageFN$ = picPage.FontName
        PageFS = picPage.FontSize
        PageFC = picPage.ForeColor
        PageFB = picPage.FontBold
        PageFT = picPage.FontTransparent
        picPage.FontName = "Courier New"
        picPage.FontBold = True
        picPage.FontTransparent = True
        For L = 1 To NLiv
            If ZLin(L).lLblFSz < 8 Then ZLin(L).lLblFSz = 8
            picPage.FontSize = ZLin(L).lLblFSz
            ' Calcolo larghezza ed altezza dell' etichetta
            ' in [ScaleUnits] ed usando il suo FontSize:
            zLbl(L).LblW = picPage.TextWidth("O") * zLbl(L).LblW
            zLbl(L).LblH = Abs(picPage.TextHeight("O"))
        Next L
    End If
'
    For L = 1 To NLiv
        If ZLin(L).lLinSps < 1 Then ZLin(L).lLinSps = 1
        picPage.DrawWidth = ZLin(L).lLinSps
        For N = 1 To NT
            ' Trovo le intersezioni del triangolo N
            ' con il piano a z = ZLin(L).dLinVal:
            plane_imp_triangle_int_3d XD(TIL(1, N)), YD(TIL(1, N)), ZD(TIL(1, N)), _
                                      XD(TIL(2, N)), YD(TIL(2, N)), ZD(TIL(2, N)), _
                                      XD(TIL(3, N)), YD(TIL(3, N)), ZD(TIL(3, N)), _
                                      0#, 0#, 1#, -ZLin(L).dLinVal, _
                                      num_int, XI(), YI(), ZI()
'
            If num_int = 1 Then
                ' Il piano interseca il triangolo N
                ' in un solo punto e lo disegno:
                picPage.PSet (XI(1), YI(1)), ZLin(L).lLinCol
            End If
'
            If num_int = 2 Then
                ' Il piano interseca il triangolo N
                ' su di un segmento e lo disegno:
                picPage.Line (XI(1), YI(1))-(XI(2), YI(2)), ZLin(L).lLinCol
                If ZLin(L).lLinSps = 1 Then picPage.PSet (XI(2), YI(2)), ZLin(L).lLinCol
'
                If fzLbl Then PreparaEtichette picPage, ZLin(), L, XI(2), YI(2)
            End If
'
'            If num_int = 3 Then
'                ' Per disegnare, anche, i triangoli giacenti
'                ' sul piano a z = ZLin(L).dLinVal:
'                picPage.Line (XI(1), YI(1))-(XI(2), YI(2)), ZLin(L).lLinCol
'                picPage.Line (XI(2), YI(2))-(XI(3), YI(3)), ZLin(L).lLinCol
'                picPage.Line (XI(3), YI(3))-(XI(1), YI(1)), ZLin(L).lLinCol
'            End If
        Next N
'
        If fStop Then GoTo DELLIV_Lines_END
    Next L
'
    ' Scrivo, se richieste, le etichette dei livelli:
    If fzLbl Then ScriviEtichette picPage, ZLin()
'
    picPage.DrawWidth = 1
    If DisegnaTriangoli Then
        ' Disegno i triangoli:
        For N = 1 To NT
            picPage.Line (XD(TIL(1, N)), YD(TIL(1, N))) _
                        -(XD(TIL(2, N)), YD(TIL(2, N))), vbWhite
            picPage.Line (XD(TIL(2, N)), YD(TIL(2, N))) _
                        -(XD(TIL(3, N)), YD(TIL(3, N))), vbWhite
            picPage.Line (XD(TIL(3, N)), YD(TIL(3, N))) _
                        -(XD(TIL(1, N)), YD(TIL(1, N))), vbWhite
        Next N
    End If
'
    If DisegnaPerimetro Then
        ' Disegno il perimetro:
        Dim LINK&, I1&
        For N = 1 To NT
            For I = 1 To 3
                LINK = TNBR(I, N)
                I1 = I + 1: If I1 > 3 Then I1 = 1
                If LINK < 0 Then
                    picPage.Line (XD(TIL(I, N)), YD(TIL(I, N))) _
                                -(XD(TIL(I1, N)), YD(TIL(I1, N))), vbWhite
                End If
            Next I
        Next N
    End If
'
'
DELLIV_Lines_END:
    picPage.DrawWidth = PageDW
    If fzLbl Then
        ' Ripristino le impostazioni:
        picPage.FontName = PageFN$
        picPage.FontSize = PageFS
        picPage.ForeColor = PageFC
        picPage.FontBold = PageFB
        picPage.FontTransparent = PageFT
    End If
'
DELLIV_Lines_ERR:
    Msg$ = IIf(Err = 0, "", Err.Description)
'
'
'
End Sub
Private Sub DisegnaTessera(ByVal picPage As PictureBox, XV#(), YV#(), ZV#(), _
    ZLin() As LineaLivello_Type)
'
'   Disegna, sul piano X-Y, il poligono con vertici XV(), YV()
'   con colore funzione di ZV(); i poligoni possono avere solo
'   3, 4 o 5 vertici.
'
    Dim I&, NV&, Zm#, THETA#(), FillColor&
    Dim L&, Xm#, Ym#
'
    NV = UBound(XV)
'
    ' Trovo l' altezza media
    ' dei punti dati:
    For I = 1 To NV
        Zm = Zm + ZV(I)
    Next I
    Zm = Zm / CDbl(NV)
'
    If NV > 3 Then
        ' Trovo il baricentro, nel piano
        ' X-Y, dei punti dati:
        Xm = 0#
        Ym = 0#
        For I = 1 To NV
            Xm = Xm + XV(I)
            Ym = Ym + YV(I)
        Next I
        Xm = Xm / CDbl(NV)
        Ym = Ym / CDbl(NV)
'
        ' Ordino, in senso antiorario ed
        ' intorno al baricentro, i punti dati:
        ReDim THETA(1 To NV)
        For I = 1 To NV
            THETA(I) = Atan2(YV(I) - Ym, XV(I) - Xm)
        Next I
        QuickSort3V THETA(), XV(), YV(), 1, NV, 1
    End If
'
    ' Trovo il colore appartenente
    ' all' altezza Zm:
    FillColor = ZLin(NLiv).lLinCol
    For L = 1 To NLiv - 1
        If ZLin(L).dLinVal <= Zm And Zm < ZLin(L + 1).dLinVal Then
            FillColor = ZLin(L).lLinCol
            Exit For
        End If
    Next L
'
    PolygonFill picPage, XV(), YV(), FillColor
'
'
'
End Sub
Private Sub PreparaEtichette(ByVal picPage As PictureBox, ZLin() As LineaLivello_Type, _
    ByVal L&, ByVal X#, ByVal Y#)
'
'
    Dim L1&
'
    ' Etichette delle linee di livello:
    If (zLbl(L).LblW > 0) And (Not zLbl(L).OK) Then ' Provo a scrivere il
                                                    ' valore della linea
                                                    ' di livello L alla
                                                    ' posizione X, Y.
        ' Scarto le posizioni fuori quadro:
        If (picPage.ScaleLeft + picPage.ScaleWidth < X + zLbl(L).LblW) _
        Or (Y - zLbl(L).LblH < picPage.ScaleTop + picPage.ScaleHeight) Then Exit Sub
        ' e quelle gia' impegnate:
        For L1 = 1 To NLiv
            If (L1 <> L) _
            And (zLbl(L1).X1 < X + zLbl(L).LblW) _
            And (X < zLbl(L1).X2) _
            And (zLbl(L1).Y1 < Y) _
            And (Y - zLbl(L).LblH < zLbl(L1).Y2) Then Exit Sub
        Next L1
'
        ' La posizione X, Y e' OK:
        zLbl(L).X1 = X                  ' Area
        zLbl(L).X2 = X + zLbl(L).LblW   ' impegnata
        zLbl(L).Y1 = Y - zLbl(L).LblH   ' dalla
        zLbl(L).Y2 = Y                  ' etichetta L.
        zLbl(L).OK = True               ' Livello L etichettato.
    End If
'
'
'
End Sub
Public Sub ScriviEtichette(ByVal picPage As PictureBox, ZLin() As LineaLivello_Type)
'
'   Scrive le etichette sui livelli:
'
    Dim L&
'
    For L = 1 To NLiv
        If zLbl(L).OK Then
            picPage.ForeColor = ZLin(L).lLblCol
            picPage.FontSize = ZLin(L).lLblFSz
'
            picPage.CurrentX = zLbl(L).X1
            picPage.CurrentY = zLbl(L).Y2
            picPage.Print ZLin(L).sLblTes
        End If
    Next L
'
'
'
End Sub
